#include <sta/debug/printing/printable.hpp>

#include <cinttypes>
#include <cstring>
#include <cstdio>
#include <cstdint>
#include <stdarg.h>

#include <sta/debug/assert.hpp>
#include <sta/lang.hpp>

namespace sta 
{
	void Printable::printf(const char * fmt, ...)
	{
		va_list args;
		va_start (args, fmt);

		char temp[1];
		int n = vsnprintf(temp, 1, fmt, args);

		va_start (args, fmt);
		char str[n+1];
		vsnprintf(str, n+1, fmt, args);
		STA_ASSERT(n > 0);

		println(str);
	}

    void Printable::print(char c, bool newline /* = false */)
    {
        print(&c, 1, newline);
    }

    void Printable::print(bool b, bool newline /* = false */)
    {
        print(b ? "true" : "false", newline);
    }

    void Printable::print(double d, bool newline /* = false */)
    {
        char buffer[64];
        snprintf(buffer, sizeof(buffer), "%f", d);
        print(buffer, newline);
    }

    void Printable::print(uint8_t num, IntegerBase base, bool newline /* = false */)
    {
        printBase(num, base, "%" PRIu8, sizeof(num), newline);
    }

    void Printable::print(uint16_t num, IntegerBase base, bool newline /* = false */)
    {
        printBase(num, base, "%" PRIu16, sizeof(num), newline);
    }

    void Printable::print(uint32_t num, IntegerBase base, bool newline /* = false */)
    {
        printBase(num, base, "%" PRIu32, sizeof(num), newline);
    }

    void Printable::print(const char * str, bool newline /* = false */)
    {
        print(str, strlen(str), newline);
    }
    
    void Printable::println()
    {
        print("\r\n", 2, false);
    }

    void Printable::println(char c)
    {
        print(&c, 1, true);
    }

    void Printable::println(bool b)
    {
        print(b, true);
    }

    void Printable::println(double d)
    {
        print(d, true);
    }

    void Printable::println(uint8_t num, IntegerBase base)
    {
        print(num, base, true);
    }

    void Printable::println(uint16_t num, IntegerBase base)
    {
        print(num, base, true);
    }

    void Printable::println(uint32_t num, IntegerBase base)
    {
        print(num, base, true);
    }

    void Printable::println(const char * str)
    {
        println(str, strlen(str));
    }

    void Printable::println(const char * str, size_t length)
    {
        print(str, length, true);
    }

    void Printable::printBase(uintmax_t num, IntegerBase base, const char * fmt, size_t size, bool newline /* = false */)
    {
        switch (base)
        {
        case IntegerBase::DEC:
            printDec(num, fmt, newline);
            break;

        case IntegerBase::BIN:
            // Digits in base 2 = size in bytes * 8
            printBin(num, size * 8, newline);
            break;

        case IntegerBase::HEX:
            // Digits in base 16 = size in bytes * 2
            printHex(num, size * 2, newline);
            break;

        default:
            print("<invalid_base>");
        }
    }

    void Printable::printDec(uintmax_t num, const char * fmt, bool newline /* = false */)
    {
        char buffer[64];
        snprintf(buffer, sizeof(buffer), fmt, static_cast<uint32_t>(num));
        print(buffer, newline);
    }

    void Printable::printBin(uintmax_t value, size_t digits, bool newline /* = false */)
    {
        // Need 8 digits for every byte
        char buffer[sizeof(value) * 8];

        // Check bounds
        if (digits > sizeof(buffer))
        {
            print("<bin_value_too_big>");
            return;
        }
        // Nothing to do
        if (digits == 0)
            return;

        for (size_t i = 0; i < digits; ++i)
        {
            // Convert bit to '0' or '1'
            // First digit in buffer is MSB in value, so shift from high to low
            buffer[i] = '0' + ((value >> (digits - 1 - i)) & 0x1);
        }

        print(buffer, digits, newline);
    }

    void Printable::printHex(uintmax_t value, size_t digits, bool newline /* = false */)
    {
        // Need 2 digits for every byte
        char buffer[sizeof(value) * 2];

        // Check bounds
        if (digits > sizeof(buffer))
        {
            print("<hex_value_too_big>", newline);
            return;
        }
        // Nothing to do
        if (digits == 0)
            return;

        for (size_t i = 0; i < digits; ++i)
        {
            // Convert 4 bits to hex
            // First digit in buffer is 4 MSBs in value, so shift from high to low
            uint8_t hex = ((value >> ((digits - 1 - i) * 4)) & 0xF);
            if (hex > 9)
                buffer[i] = 'A' + (hex - 10);
            else
                buffer[i] = '0' + hex;
        }

        print(buffer, digits, newline);
    }

    void Printable::read(char * str, size_t length)
    {
        read(reinterpret_cast<uint8_t*>(str), length * sizeof(char));
    }

    void Printable::read(float * buffer, size_t length)
    {
        read(reinterpret_cast<uint8_t*>(buffer), length * sizeof(float));
    }

    void Printable::read(double * buffer, size_t length)
    {
        read(reinterpret_cast<uint8_t*>(buffer), length * sizeof(double));
    }

} // namespace sta
