#include <sta/devices/stm32/bus/i2c.hpp>

#include <sta/debug/assert.hpp>
#include <cstring>

#ifdef STA_STM32_I2C_ENABLED

namespace sta
{
	STM32I2C::STM32I2C(I2C_HandleTypeDef * handle, Mutex * mutex)
			: I2C{mutex}, handle_{handle}
	{
		STA_ASSERT(handle != nullptr);
	}

	bool STM32I2C::transfer(uint8_t value, uint32_t timeout /* = STA_MAX_TIMEOUT */)
	{
		if (master_)
		{
			return HAL_I2C_Master_Transmit(handle_, address_, &value, 1, timeout) == HAL_OK;
		} else {
			return HAL_I2C_Slave_Transmit(handle_, &value, 1, timeout) == HAL_OK;
		}
	}

	bool STM32I2C::transfer16(uint16_t value, uint32_t timeout /* = STA_MAX_TIMEOUT */)
	{
		if (blocking_)
		{
			if (master_)
			{
				return HAL_I2C_Master_Transmit(handle_, address_, reinterpret_cast<uint8_t *>(&value), 2, timeout) == HAL_OK;
			} else {
				return HAL_I2C_Slave_Transmit(handle_, reinterpret_cast<uint8_t *>(&value), 2, timeout) == HAL_OK;
			}
		} else {
			if (master_)
			{
				return HAL_I2C_Slave_Transmit_IT(handle_, reinterpret_cast<uint8_t *>(&value), 2) == HAL_OK;
			} else {
				return HAL_I2C_Slave_Transmit_IT(handle_, reinterpret_cast<uint8_t *>(&value), 2) == HAL_OK;
			}
		}
	}

	bool STM32I2C::transfer(const uint8_t * buffer, size_t size, uint32_t timeout /* = STA_MAX_TIMEOUT */)
	{
		HAL_StatusTypeDef res;

		/*
		 *  It's undecided if we want to change the parameter for this function. Since the transmission
		 *  doesn't take a const buffer as an argument, we are using this fix by creating a temporary buffer.
		 */
		uint8_t * temp_buffer = new uint8_t[size];
		memcpy(temp_buffer, buffer, size);

		if (blocking_)
		{
			if (master_)
			{
				res = HAL_I2C_Master_Transmit(handle_, address_, temp_buffer, size, timeout);
			} else {
				res = HAL_I2C_Slave_Transmit(handle_, temp_buffer, size, timeout);
			}
		} else {
			if (master_) {
				res = HAL_I2C_Master_Transmit_IT(handle_, address_, temp_buffer, size);
			} else {
				res = HAL_I2C_Slave_Transmit_IT(handle_, temp_buffer, size);
			}
		}

		delete [] temp_buffer;

		return res == HAL_OK;
	}

	bool STM32I2C::writeMemory(uint8_t regAddr, const uint8_t * buffer, size_t size, uint32_t timeout /* = STA_MAX_TIMEOUT */)
	{
		/*
		 *  It's undecided if we want to change the parameter for this function. Since the transmission
		 *  doesn't take a const buffer as an argument, we are using this fix by creating a temporary buffer.
		 */
		uint8_t * temp_buffer = new uint8_t[size];
		memcpy(temp_buffer, buffer, size);

		if (blocking_) {
			return HAL_I2C_Mem_Write(handle_, address_, regAddr, 1, temp_buffer, size, timeout) == HAL_OK;
		} else {
			return HAL_I2C_Mem_Write_IT(handle_, address_, regAddr, 1, temp_buffer, size) == HAL_OK;
		}
	}

	bool STM32I2C::receive(uint8_t * buffer, size_t size, uint32_t timeout /* = STA_MAX_TIMEOUT */)
	{
		if (blocking_) {
			if (master_) {
				return HAL_I2C_Master_Receive(handle_, address_, buffer, size, timeout) == HAL_OK;
			} else {
				return HAL_I2C_Slave_Receive(handle_, buffer, size, timeout) == HAL_OK;
			}
		} else {
			if (master_) {
				return HAL_I2C_Master_Receive_IT(handle_, address_, buffer, size) == HAL_OK;
			} else {
				return HAL_I2C_Slave_Receive_IT(handle_, buffer, size) == HAL_OK;
			}
		}
	}

	bool STM32I2C::readMemory(uint8_t regAddr, uint8_t * buffer, size_t size, uint32_t timeout /* = STA_MAX_TIMEOUT */)
	{
		if (blocking_) {
			return HAL_I2C_Mem_Read(handle_, address_, regAddr, 1, buffer, size, timeout) == HAL_OK;
		} else {
			return HAL_I2C_Mem_Read_IT(handle_, address_, regAddr, 1, buffer, size) == HAL_OK;
		}
	}

	bool STM32I2C::fill(uint8_t value, size_t count, uint32_t timeout /* = STA_MAX_TIMEOUT */)
	{
		// Initialize a buffer of size count and fill it with the value.
		uint8_t *buffer = new uint8_t[count];
		memset(buffer, value, count);

		// Transfer the buffer via the bus.
		bool rslt = transfer(buffer, count, timeout);

		delete [] buffer;

		return rslt;
	}

	STM32I2CDevice::STM32I2CDevice(STM32I2C * intf, int address, bool master, bool blocking)
		/* The address is bit-shifted by one to work properly for the STM32 HAL. */
		: I2CDevice(intf, address << 1, master, blocking)
	{

	}

} // namespace sta

#endif // STA_STM32_I2C_ENABLED
